// ============================================================================ //
// Lightflow Rendering Interface                                                //
// ============================================================================ //
// LfMatrix3.h                                                                  //
// ============================================================================ //
//                                                                              //
// The contents of this file are subject to the Lightflow Rendering Interface   //
// License (the License); you may not use this file except in compliance with   //
// the License. You may obtain a copy of the License by sending an email at     //
// jp@lightflowtech.com with subject LRI-LICENSE.                               //
//                                                                              //
// ============================================================================ //
// Copyright (c) 1995-2000 by Marco and Jacopo Pantaleoni. All rights reserved. //
// ============================================================================ //

#ifndef _LfMatrix3_h
#define _LfMatrix3_h 1

#include <Lightflow/LfVector3.h>


class LfAPI_CS LfMatrix3
{
private:
	LfVector3 r[3];

public:
inline                      LfMatrix3   ();
inline                      LfMatrix3   (const LfMatrix3&);
inline                      LfMatrix3   (const LfVector3&, const LfVector3&, const LfVector3&);
inline                      LfMatrix3   (const float  v[3][3]);
inline                      LfMatrix3   (const double v[3][3]);

inline        LfMatrix3&    operator  = (const LfMatrix3&);
inline        LfMatrix3&    operator += (const LfMatrix3&);
inline        LfMatrix3&    operator -= (const LfMatrix3&);
inline        LfMatrix3&    operator *= (const LfMatrix3&);
inline        LfMatrix3&    operator *= (LfFloat);
inline        LfMatrix3&    operator /= (LfFloat);

inline        LfVector3&       operator [] (int);
inline        const LfVector3& Get(int) const;
inline        void             Set(int, const LfVector3&);

inline        LfFloat       Det() const;

friend inline LfAPI_CS int           operator == (const LfMatrix3&,  const LfMatrix3&);
friend inline LfAPI_CS int           operator != (const LfMatrix3&,  const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3     operator  - (const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3     operator  + (const LfMatrix3&,  const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3     operator  - (const LfMatrix3&,  const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3     operator  * (const LfMatrix3&,  const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3     operator  * (const LfMatrix3&,  LfFloat);
friend inline LfAPI_CS LfMatrix3     operator  * (LfFloat,           const LfMatrix3&);
friend inline LfAPI_CS LfVector3     operator  * (const LfVector3&,  const LfMatrix3&);
friend inline LfAPI_CS LfVector3     operator  * (const LfMatrix3&,  const LfVector3&);
friend inline LfAPI_CS LfMatrix3&    Multiply    (const LfMatrix3&,  const LfMatrix3&,  LfMatrix3&);
friend inline LfAPI_CS LfVector3&    Multiply    (const LfVector3&,  const LfMatrix3&,  LfVector3&);
friend inline LfAPI_CS LfVector3&    Multiply    (const LfMatrix3&,  const LfVector3&,  LfVector3&);
friend inline LfAPI_CS LfMatrix3     operator  / (const LfMatrix3&,  LfFloat);
friend inline LfAPI_CS LfMatrix3     Transpose   (const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3&    Transpose   (const LfMatrix3&,  LfMatrix3&);
/*friend inline LfAPI_CS LfMatrix3   Invert      (const LfMatrix3&);
friend inline LfAPI_CS LfMatrix3     Invert      (const LfMatrix3&,  int&);*/
friend inline LfAPI_CS int           Invert      (const LfMatrix3&,  LfMatrix3&);
friend inline LfAPI_CS LfFloat       Det         (const LfMatrix3&);
};


//
// I M P L E M E N T A T I O N
//

#ifdef _V
#undef _V
#endif
#define _V(v, i)   (((LfVector3&)v).x[(i)])

#ifdef _M
#undef _M
#endif
#define _M(m, i, j)   (((LfMatrix3&)m).r[(i)][(j)])

#ifdef _CM
#undef _CM
#endif
#define _CM(m, i, j)  ((const LfFloat&)(((LfMatrix3&)m).r[(i)][(j)]))


//
// LfMatrix3 inline methods
//

inline LfMatrix3::LfMatrix3() { }

inline LfMatrix3::LfMatrix3(const LfMatrix3& m)
{
	r[0] = m.r[0];
	r[1] = m.r[1];
	r[2] = m.r[2];
}

inline LfMatrix3::LfMatrix3(const LfVector3& r0, const LfVector3& r1, const LfVector3& r2)
{
	r[0] = r0;
	r[1] = r1;
	r[2] = r2;
}

inline LfMatrix3::LfMatrix3(const float  v[3][3])
{
	r[0][0] = v[0][0];
	r[0][1] = v[0][1];
	r[0][2] = v[0][2];

	r[1][0] = v[1][0];
	r[1][1] = v[1][1];
	r[1][2] = v[1][2];

	r[2][0] = v[2][0];
	r[2][1] = v[2][1];
	r[2][2] = v[2][2];
}

inline LfMatrix3::LfMatrix3(const double v[3][3])
{
	r[0][0] = v[0][0];
	r[0][1] = v[0][1];
	r[0][2] = v[0][2];

	r[1][0] = v[1][0];
	r[1][1] = v[1][1];
	r[1][2] = v[1][2];

	r[2][0] = v[2][0];
	r[2][1] = v[2][1];
	r[2][2] = v[2][2];
}

inline LfMatrix3& LfMatrix3::operator  = (const LfMatrix3& m)
{
	r[0] = m.r[0];
	r[1] = m.r[1];
	r[2] = m.r[2];

	return *this;
}

inline LfMatrix3& LfMatrix3::operator += (const LfMatrix3& m)
{
	r[0] += m.r[0];
	r[1] += m.r[1];
	r[2] += m.r[2];

	return *this;
}

inline LfMatrix3& LfMatrix3::operator -= (const LfMatrix3& m)
{
	r[0] -= m.r[0];
	r[1] -= m.r[1];
	r[2] -= m.r[2];

	return *this;
}

// SUBOPTIMAL
inline LfMatrix3& LfMatrix3::operator *= (const LfMatrix3& m)
// WARNING: This routine uses two temporary variables, so, if you need
//          to multiply two matrices two obtain a third matrix, AVOID
//          this method. Use instead:
//            LfMatrix3& Multiply(const LfMatrix3& m1, const LfMatrix3& m2, LfMatrix3& r)
{
	LfFloat t1 = r[0][0];
	LfFloat t2 = r[0][1];

	r[0][0] = r[0][0]*_M(m,0,0) + r[0][1]*_M(m,1,0) + r[0][2]*_M(m,2,0);
	r[0][1] = t1     *_M(m,0,1) + r[0][1]*_M(m,1,1) + r[0][2]*_M(m,2,1);
	r[0][2] = t1     *_M(m,0,2) + t2     *_M(m,1,2) + r[0][2]*_M(m,2,2);


	t1 = r[1][0];
	t2 = r[1][1];

	r[1][0] = r[1][0]*_M(m,0,0) + r[1][1]*_M(m,1,0) + r[1][2]*_M(m,2,0);
	r[1][1] = t1     *_M(m,0,1) + r[1][1]*_M(m,1,1) + r[1][2]*_M(m,2,1);
	r[1][2] = t1     *_M(m,0,2) + t2     *_M(m,1,2) + r[1][2]*_M(m,2,2);


	t1 = r[2][0];
	t2 = r[2][1];

	r[2][0] = r[2][0]*_M(m,0,0) + r[2][1]*_M(m,1,0) + r[2][2]*_M(m,2,0);
	r[2][1] = t1     *_M(m,0,1) + r[2][1]*_M(m,1,1) + r[2][2]*_M(m,2,1);
	r[2][2] = t1     *_M(m,0,2) + t2     *_M(m,1,2) + r[2][2]*_M(m,2,2);

	return *this;
}

inline LfMatrix3& LfMatrix3::operator *= (LfFloat k)
{
	r[0] *= k;
	r[1] *= k;
	r[2] *= k;

	return *this;
}

inline LfMatrix3& LfMatrix3::operator /= (LfFloat k)
{
	r[0] /= k;
	r[1] /= k;
	r[2] /= k;

	return *this;
}

inline LfVector3& LfMatrix3::operator [] (int i)
{
	return r[i];
}
inline const LfVector3& LfMatrix3::Get(int i) const
{
	return r[i];
}
inline void LfMatrix3::Set(int i, const LfVector3& v)
{
	r[i] = v;
}

inline LfFloat LfMatrix3::Det() const
{
	return (_CM(*this,0,0)*(_CM(*this,1,1)*_CM(*this,2,2)-_CM(*this,1,2)*_CM(*this,2,1))+
			_CM(*this,0,1)*(_CM(*this,1,2)*_CM(*this,2,0)-_CM(*this,1,0)*_CM(*this,2,2))+
			_CM(*this,0,2)*(_CM(*this,1,0)*_CM(*this,2,1)-_CM(*this,1,1)*_CM(*this,2,0)));
}

//
// LfMatrix3 inline functions (not members)
//

inline LfAPI_CS int operator == (const LfMatrix3& a, const LfMatrix3& b)
{
	return ((a.r[0] == b.r[0]) && (a.r[1] == b.r[1]) && (a.r[2] == b.r[2]));
}

inline LfAPI_CS int operator != (const LfMatrix3& a, const LfMatrix3& b)
{
	return !(a == b);
}

inline LfAPI_CS LfMatrix3 operator  - (const LfMatrix3& a)
{
	return (LfMatrix3(a) *= -1.0);
}

inline LfAPI_CS LfMatrix3 operator  + (const LfMatrix3& a, const LfMatrix3& b)
{
	LfMatrix3 r;
	r.r[0][0] = _M(a,0,0) + _M(b,0,0);
	r.r[0][1] = _M(a,0,1) + _M(b,0,1);
	r.r[0][2] = _M(a,0,2) + _M(b,0,2);
	r.r[1][0] = _M(a,1,0) + _M(b,1,0);
	r.r[1][1] = _M(a,1,1) + _M(b,1,1);
	r.r[1][2] = _M(a,1,2) + _M(b,1,2);
	r.r[2][0] = _M(a,2,0) + _M(b,2,0);
	r.r[2][1] = _M(a,2,1) + _M(b,2,1);
	r.r[2][2] = _M(a,2,2) + _M(b,2,2);
	return r;
}

inline LfAPI_CS LfMatrix3 operator  - (const LfMatrix3& a, const LfMatrix3& b)
{
	LfMatrix3 r;
	r.r[0][0] = _M(a,0,0) - _M(b,0,0);
	r.r[0][1] = _M(a,0,1) - _M(b,0,1);
	r.r[0][2] = _M(a,0,2) - _M(b,0,2);
	r.r[1][0] = _M(a,1,0) - _M(b,1,0);
	r.r[1][1] = _M(a,1,1) - _M(b,1,1);
	r.r[1][2] = _M(a,1,2) - _M(b,1,2);
	r.r[2][0] = _M(a,2,0) - _M(b,2,0);
	r.r[2][1] = _M(a,2,1) - _M(b,2,1);
	r.r[2][2] = _M(a,2,2) - _M(b,2,2);

	return r;
}

inline LfAPI_CS LfMatrix3 operator  * (const LfMatrix3& a, const LfMatrix3& b)
{
	LfMatrix3 r;

	return Multiply(a, b, r);
}

inline LfAPI_CS LfMatrix3 operator  * (const LfMatrix3& a, LfFloat k)
{
	LfMatrix3 r;
	r.r[0][0] = _M(a,0,0) * k;
	r.r[0][1] = _M(a,0,1) * k;
	r.r[0][2] = _M(a,0,2) * k;
	r.r[1][0] = _M(a,1,0) * k;
	r.r[1][1] = _M(a,1,1) * k;
	r.r[1][2] = _M(a,1,2) * k;
	r.r[2][0] = _M(a,2,0) * k;
	r.r[2][1] = _M(a,2,1) * k;
	r.r[2][2] = _M(a,2,2) * k;

	return r;
}

inline LfAPI_CS LfMatrix3 operator  * (LfFloat k, const LfMatrix3& a)
{
	return (a * k);
}

inline LfAPI_CS LfVector3 operator * (const LfVector3& v, const LfMatrix3& m)
{
	LfVector3 r;

	return Multiply(v, m, r);
}

inline LfAPI_CS LfVector3 operator * (const LfMatrix3& m, const LfVector3& v)
{
	LfVector3 r;

	return Multiply(m, v, r);
}

inline LfAPI_CS LfMatrix3& Multiply(const LfMatrix3& a, const LfMatrix3& b, LfMatrix3& r)
{
	r.r[0][0] = _M(a,0,0)*_M(b,0,0) + _M(a,0,1)*_M(b,1,0) + _M(a,0,2)*_M(b,2,0);
	r.r[0][1] = _M(a,0,0)*_M(b,0,1) + _M(a,0,1)*_M(b,1,1) + _M(a,0,2)*_M(b,2,1);
	r.r[0][2] = _M(a,0,0)*_M(b,0,2) + _M(a,0,1)*_M(b,1,2) + _M(a,0,2)*_M(b,2,2);

	r.r[1][0] = _M(a,1,0)*_M(b,0,0) + _M(a,1,1)*_M(b,1,0) + _M(a,1,2)*_M(b,2,0);
	r.r[1][1] = _M(a,1,0)*_M(b,0,1) + _M(a,1,1)*_M(b,1,1) + _M(a,1,2)*_M(b,2,1);
	r.r[1][2] = _M(a,1,0)*_M(b,0,2) + _M(a,1,1)*_M(b,1,2) + _M(a,1,2)*_M(b,2,2);

	r.r[2][0] = _M(a,2,0)*_M(b,0,0) + _M(a,2,1)*_M(b,1,0) + _M(a,2,2)*_M(b,2,0);
	r.r[2][1] = _M(a,2,0)*_M(b,0,1) + _M(a,2,1)*_M(b,1,1) + _M(a,2,2)*_M(b,2,1);
	r.r[2][2] = _M(a,2,0)*_M(b,0,2) + _M(a,2,1)*_M(b,1,2) + _M(a,2,2)*_M(b,2,2);

	return r;
}

// OPTIMAL
inline LfAPI_CS LfVector3& Multiply(const LfVector3& v, const LfMatrix3& m, LfVector3& r)
{
	r[0] = _V(v,0)*_M(m,0,0) + _V(v,1)*_M(m,1,0) + _V(v,2)*_M(m,2,0);
	r[1] = _V(v,0)*_M(m,0,1) + _V(v,1)*_M(m,1,1) + _V(v,2)*_M(m,2,1);
	r[2] = _V(v,0)*_M(m,0,2) + _V(v,1)*_M(m,1,2) + _V(v,2)*_M(m,2,2);

	return r;
}

// OPTIMAL
inline LfAPI_CS LfVector3& Multiply(const LfMatrix3& m, const LfVector3& v, LfVector3& r)
{
	r[0] = _M(m,0,0)*_V(v,0) + _M(m,0,1)*_V(v,1) + _M(m,0,2)*_V(v,2);
	r[1] = _M(m,1,0)*_V(v,0) + _M(m,1,1)*_V(v,1) + _M(m,1,2)*_V(v,2);
	r[2] = _M(m,2,0)*_V(v,0) + _M(m,2,1)*_V(v,1) + _M(m,2,2)*_V(v,2);

	return r;
}

inline LfAPI_CS LfMatrix3 operator  / (const LfMatrix3& a, LfFloat k)
{
	LfMatrix3 r;
	r.r[0][0] = _M(a,0,0) / k;
	r.r[0][1] = _M(a,0,1) / k;
	r.r[0][2] = _M(a,0,2) / k;
	r.r[1][0] = _M(a,1,0) / k;
	r.r[1][1] = _M(a,1,1) / k;
	r.r[1][2] = _M(a,1,2) / k;
	r.r[2][0] = _M(a,2,0) / k;
	r.r[2][1] = _M(a,2,1) / k;
	r.r[2][2] = _M(a,2,2) / k;

	return r;
}

inline LfAPI_CS LfMatrix3 Transpose(const LfMatrix3& m)
{
	LfMatrix3 r;

	return Transpose(m, r);
}

inline LfAPI_CS LfMatrix3& Transpose(const LfMatrix3& m, LfMatrix3& r)
{
	r.r[0][0] = _M(m,0,0);
	r.r[0][1] = _M(m,1,0);
	r.r[0][2] = _M(m,2,0);
	r.r[1][0] = _M(m,0,1);
	r.r[1][1] = _M(m,1,1);
	r.r[1][2] = _M(m,2,1);
	r.r[2][0] = _M(m,0,2);
	r.r[2][1] = _M(m,1,2);
	r.r[2][2] = _M(m,2,2);

	return r;
}

inline LfAPI_CS int Invert(const LfMatrix3& a, LfMatrix3& r)
{
	LfFloat d;

	// We calculate the adjoint matrix of a, then we divide it by the determinant of a

	r.r[0][0] = _M(a,1,1)*_M(a,2,2) - _M(a,1,2)*_M(a,2,1);
	r.r[0][1] = _M(a,0,2)*_M(a,2,1) - _M(a,0,1)*_M(a,2,2);
	r.r[0][2] = _M(a,0,1)*_M(a,1,2) - _M(a,0,2)*_M(a,1,1);

	r.r[1][0] = _M(a,1,2)*_M(a,2,0) - _M(a,1,0)*_M(a,2,2);
	r.r[1][1] = _M(a,0,0)*_M(a,2,2) - _M(a,0,2)*_M(a,2,0);
	r.r[1][2] = _M(a,0,2)*_M(a,1,0) - _M(a,0,0)*_M(a,1,2);

	r.r[2][0] = _M(a,1,0)*_M(a,2,1) - _M(a,1,1)*_M(a,2,0);
	r.r[2][1] = _M(a,0,1)*_M(a,2,0) - _M(a,0,0)*_M(a,2,1);
	r.r[2][2] = _M(a,0,0)*_M(a,1,1) - _M(a,0,1)*_M(a,1,0);

	// We save some time calculating Det(a) this way (now r is adjoint of a)
	d = _M(a,0,0)*r.r[0][0] + _M(a,0,1)*r.r[1][0] + _M(a,0,2)*r.r[2][0];

	if (__LfZero__(d)) return 0; // Singular matrix => no inverse

	d = 1/d;

	r *= d;

	return 1;
}

inline LfAPI_CS LfFloat Det(const LfMatrix3& m)
{
	return m.Det();
}

#undef _M
#undef _CM

#endif /* _LfMatrix3_h */
