// ============================================================================ //
// Lightflow Rendering Interface                                                //
// ============================================================================ //
// LfThreadSaver.h                                                              //
// ============================================================================ //
//                                                                              //
// The contents of this file are subject to the Lightflow Rendering Interface   //
// License (the License); you may not use this file except in compliance with   //
// the License. You may obtain a copy of the License by sending an email at     //
// jp@lightflowtech.com with subject LRI-LICENSE.                               //
//                                                                              //
// ============================================================================ //
// Copyright (c) 1995-2000 by Jacopo Pantaleoni. All rights reserved.           //
// ============================================================================ //

#ifndef _LfThreadSaver_H
#define _LfThreadSaver_H 1

#include <Lightflow/LfBasicTypes.h>


#if LfPTHREADS

typedef pthread_t LfThread;
typedef void*     LfThreadData;
typedef void* (*LfThreadFunction)( LfThreadData );

#define LfTHREADRETURN void*

#elif LfWINTHREADS

typedef HANDLE LfThread;
typedef LPVOID LfThreadData;
typedef UINT (*LfThreadFunction)( LfThreadData );

#define LfTHREADRETURN UINT WINAPI

#else

typedef LfInt LfThread;
typedef void* LfThreadData;
typedef void* (*LfThreadFunction)( LfThreadData );

#define LfTHREADRETURN void*

#endif


class LfAPI_CS LfMutex
{
private:
#if LfPTHREADS
	pthread_mutex_t mutex;
#elif LfWINTHREADS
	CRITICAL_SECTION mutex;
#endif

public:
	LfMutex( void );
	LfMutex( LfMutexAttribute& attr );
	~LfMutex();

	LfInt Lock    (void);
	LfInt TryLock (void);
	LfInt UnLock  (void);
};

class LfAPI_CS LfMutexAttribute
{
private:
#if LfPTHREADS
	pthread_mutexattr_t mutexattr;
#endif

public:
	LfMutexAttribute(void);
	~LfMutexAttribute();

	LfInt SetKind (LfInt  kind);
	LfInt GetKind (LfInt& kind);

	friend LfMutex::LfMutex( LfMutexAttribute& attr );
};


LfAPI_CS LfInt LfThreadCreate ( LfThread* thread, LfThreadFunction process, LfThreadData data );
LfAPI_CS LfInt LfThreadJoin   ( LfThread  thread, void** retdata );


#if LfGNU
	extern LfMutex LfStreamLocker;

	#define LfStream(format, args...)    \
	do {                                 \
	  LfStreamLocker.Lock();             \
	  fprintf(stderr, format, ## args);  \
	  fflush(stderr);                    \
	  LfStreamLocker.UnLock();           \
	} while(0)
#else
	#define LfStream printf
#endif

#endif
